# This file was auto-generated by Fern from our API Definition.

import enum
import typing

T_Result = typing.TypeVar("T_Result")


class MultiSeriesInsampleForecastModel(str, enum.Enum):
    """
    Model to use as a string. Options are: `timegpt-1`, and `timegpt-1-long-horizon.` We recommend using `timegpt-1-long-horizon` for forecasting if you want to predict more than one seasonal period given the frequency of your data.
    """

    TIMEGPT_1 = "timegpt-1"
    TIMEGPT_1_LONG_HORIZON = "timegpt-1-long-horizon"

    def visit(
        self, timegpt_1: typing.Callable[[], T_Result], timegpt_1_long_horizon: typing.Callable[[], T_Result]
    ) -> T_Result:
        if self is MultiSeriesInsampleForecastModel.TIMEGPT_1:
            return timegpt_1()
        if self is MultiSeriesInsampleForecastModel.TIMEGPT_1_LONG_HORIZON:
            return timegpt_1_long_horizon()
